/* Trux's Tunneling Protocol */

#include <arpa/inet.h>  /* needed for inet_aton() and inet_addr() */
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/stropts.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>
#include <math.h>

#include "diffie_hellman.c"
#include "circular_buffer.c"

#define STATE_BUFFER_SIZE 262144
/* #define STATE_BUFFER_SIZE 32768 -- causes problems... */
#define MAX_TTP_PACKET_SIZE 1200  /* be careful not to exceed underlying transport's MTU */
#define MAX_TCP_CONNECTIONS 8  /* per server port */
#define MAX_NUM_INTERFACES 1024  /* per ttp process */
#define RETRANSMIT_TIMEOUT_USEC 809749  /* time since oldest packet before requesting a retransmit */
#define RETRANSMIT_POLL_PERIOUD_USEC 1600453  /* period of repeated retransmit requests */
#define ABANDON_TIMEOUT_USEC 16000453  /* time since oldest packet before abandoning missing data */
#define REPORT_CAPACITY_PERIOD_USEC 32000287  /* period between reporting channel capacity */
#define RECALCULATE_WEIGHTS_PERIOD_USEC 809001  /* period between recalculating weights */
#define PING_PERIOD_USEC 8000238  /* period between sending pings */
#define PING_TIMEOUT_USEC 28000255  /* time since last ping before putting interface on standby */
#define DEATH_TIMEOUT_USEC 128000000  /* time since standby until death is signaled */
#define STATUS_LINE_PERIOD_USEC 16000183  /* interval between printing status lines */
#define TTP_HEADER_SIZE (6*sizeof(int))  /* magic, checksum, size, position, create_sec, create_usec */
#define TTP_CORE_HEADER_SIZE (4*sizeof(int))  /* deterministic fields, excluding timestamp */
#define ROLLOVER 10281971

#define HISTORY_LENGTH 32
#define MAX_HISTORY_TIME 8.0
#define UNDERUTILIZATION_THRESHHOLD 0.25  /* less than 25% of capacity is considered underutilized */

#define boolean int

#define TTP_MAGIC 0x10281971
#define TTP_ACTIVE 0x0DE1FE1F
#define TTP_INACTIVE 0xDEADBEEF
#define TTP_PING 0x05155150
#define TTP_RETRANSMIT 0xD00B0809
#define TTP_FEEDBACK 0x83FABC38

int TTP_DEBUG=1 ;
int TTP_KEEP_DEBUG_LOGS=0 ;

char *resource_filename ;
char history_buffer[2][STATE_BUFFER_SIZE] ;
char side_buffer[2][STATE_BUFFER_SIZE] ;
char side_flags[2][STATE_BUFFER_SIZE] ;  /* indicates which bytes in the buffer are valid */
int side_counter[2] ;  /* marks the index of side_buffer[x][0] as the buffer is shifted out */
int side_total[2] ;  /* counts the total number of valid flags (between 0 and STATE_BUFFER_SIZE) */
long total_abandoned[2] ;  /* counts the total number of bytes abandoned */
struct timeval oldest_packet_time[2] ;  /* timestamp of oldest packet sitting in queue */
struct timeval last_retransmit_time[2] ;  /* when did we last request a retransmit? */
int oldest_packet_position[2] ;  /* position of oldest packet sitting in queue */
boolean ttp_is_relay ;
int reload_flag ;
int main_ttp_pid ;
char pppdname[512] ;

#define VALID_FLAG 1
#define RETRANSMIT_FLAG 2

enum interface_type
{
  ttp_ppp,
  ttp_tty,
  ttp_tcp_transient,  /* transient tcp connection after being accepted by server */
  ttp_tcp_client,  /* client tcp connection */
  ttp_udp
} ;

struct connection_state
{
  /* this data is used by ttp and its interfaces */
  char *name ;
  enum interface_type interface ;
  int ttp_to_interface[2] ;  /* pipe */
  int interface_to_ttp[2] ;  /* pipe */
  int ttp_death_signal[2] ;  /* pipe */

  int pid ;  /* an interface forks itself */

  /* this data is used by ttp only */
  int flag ;  /* temporary flag for searches */
  boolean side ;  /* whether the interface is on the left or the right of the equals sign */
  boolean raw ; /* endpoint like ppp or tty */
  boolean dead ; /* is the interface dead? */
  boolean active ; /* is the interface connected? */
  boolean standby ;  /* is the interface having problems (or reconnecting) right now? */
  long total_bytes_in ;
  long total_bytes_out ;
  long total_too_late ;
  int send_packet_history_index ;  /* counter that increments % HISTORY_LENGTH */
  long send_packet_length_history[HISTORY_LENGTH] ;
  struct timeval send_packet_time_history[HISTORY_LENGTH] ;
  int receive_packet_history_index ;  /* counter that increments % HISTORY_LENGTH */
  long receive_packet_length_history[HISTORY_LENGTH] ;
  struct timeval receive_packet_time_history[HISTORY_LENGTH] ;
  int destination_flag ;  /* temporary flag, assigned at the most recent call to disburse_data() */
  double weight ;  /* temporary variable, assigned at the most recent call to disburse_data() */
  double send_utilization ;  /* bytes per second in the last 4 seconds */
  double receive_utilization ;  /* bytes per second in the last 4 seconds */
  double send_capacity ;  /* max bytes ber second */
  double receive_capacity ;  /* max bytes ber second */
  struct timeval initialization_time ;
  struct timeval newest_packet_time ;  /* when did we last get a packet over this interface? */
  struct timeval last_send_time ;  /* when did we last send a ping over this interface? */
  struct timeval last_report_capacity_time ;  /* when did we last report capacity over this interface? */
  struct circular_buffer_type *cb ;  /* comes from interface_to_ttp[0], goes to another interface */
  struct connection_state *previous ;
  char debug_filename[256] ;
  FILE *debug_file ;
} *connections=NULL ;

char *timestamps()
{
  static char result[128] ;
  char usecs[128] ;
  struct timeval now ;

  gettimeofday(&now,NULL) ;
  sprintf(result,"%s",asctime(localtime(&(now.tv_sec)))) ;
  result[strlen(result)-1]=0 ;  /* chop off trailing newline */
  sprintf(usecs," [%06ld]",now.tv_usec) ;
  strcat(result,usecs) ;
  return(result) ;
}

void print_status_line()
{
  struct connection_state *cs ;
  static int status_line_counter=0 ;
  double overall_receive_utilization=0.0 ;
  double overall_send_utilization=0.0 ;

  if(status_line_counter%1==0)
    {
      printf("\n") ;
      for(cs=connections;cs!=NULL;cs=cs->previous)
	if(!cs->raw)
	  printf("  %24s  ",cs->name) ;
      printf("\n") ;
    }
  for(cs=connections;cs!=NULL;cs=cs->previous)
    if(!cs->raw)
      {
	printf("  s%0.0lf/%0.0lf:r%0.0lf/%0.0lf=%0.0lf  ",cs->send_utilization,cs->send_capacity,cs->receive_utilization,cs->receive_capacity,cs->weight) ;
	overall_receive_utilization+=cs->receive_utilization ;
	overall_send_utilization+=cs->send_utilization ;
      }
  status_line_counter++ ;
  printf("\n                                  (aggregate rcv %0.0lf, snd %0.0lf)\n",overall_receive_utilization,overall_send_utilization) ;
}

int send_data_to_interface(struct connection_state *destination, char *buf, int buflen, int position, int magic) ;  /* prototype */

/* tell other side about its send_capacity (our receive_capacity) */
void report_remote_capacity(struct connection_state *cs)
{
  char buf[128] ;
  int buflen=0 ;
  int position=0 ;
  int magic=TTP_FEEDBACK ;

  memcpy(&(buf[buflen]),&(cs->receive_capacity),sizeof(double)) ;
  buflen+=sizeof(double) ;

  send_data_to_interface(cs, buf, buflen, position, magic) ;
}

struct timeval last_recalc_time ;

void recalculate_weights(int verbose)
{
  struct connection_state *cs ;
  double weight, delta_time, biggest_delta_time ;
  unsigned long nbytes ;
  struct timeval now ;
  long delta ;
  int i ;

  gettimeofday(&now,NULL) ;

  delta=(1000000*(now.tv_sec-last_recalc_time.tv_sec)) +
        (now.tv_usec-last_recalc_time.tv_usec) ;
  if((delta>=RECALCULATE_WEIGHTS_PERIOD_USEC) || (delta<0))
    {
      gettimeofday(&last_recalc_time,NULL) ;
      
      /* assign weights */
      for(cs=connections;cs!=NULL;cs=cs->previous)
	if((cs->active)&&(!(cs->standby)))
	  {
	    if(verbose)
	      fprintf(stderr,"\tfiguring weight for [%s] ...\n",cs->name) ;
	    
	    nbytes=0 ;
	    biggest_delta_time=0.0 ;
	    for(i=0;i<HISTORY_LENGTH;i++)
	      {
		delta_time=(double)(now.tv_sec-cs->send_packet_time_history[i].tv_sec) +
		  (0.000001*(double)(now.tv_usec-cs->send_packet_time_history[i].tv_usec)) ;
		if(delta_time<MAX_HISTORY_TIME)
		  {
		    nbytes+=cs->send_packet_length_history[i] ;
		    if(delta_time>biggest_delta_time)
		      biggest_delta_time=delta_time ;
		  }
		cs->send_utilization=0.0 ;
		if(biggest_delta_time>1.0)
		  cs->send_utilization=((double) nbytes)/(biggest_delta_time) ;
	      }
	    
	    nbytes=0 ;
	    biggest_delta_time=0.0 ;
	    for(i=0;i<HISTORY_LENGTH;i++)
	      {
		delta_time=(double)(now.tv_sec-cs->receive_packet_time_history[i].tv_sec) +
		  (0.000001*(double)(now.tv_usec-cs->receive_packet_time_history[i].tv_usec)) ;
		if(delta_time<MAX_HISTORY_TIME)
		  {
		    nbytes+=cs->receive_packet_length_history[i] ;
		    if(delta_time>biggest_delta_time)
		      biggest_delta_time=delta_time ;
		  }
		cs->receive_utilization=0.0 ;
		if(biggest_delta_time>1.0)
		  cs->receive_utilization=((double) nbytes)/(biggest_delta_time) ;
		if(cs->receive_utilization > cs->receive_capacity)
		  cs->receive_capacity=cs->receive_utilization ;
		/* tell other side about its send_capacity (our receive_capacity) */
		/* do this periodically, because the transport may be unreliable */
		if(!cs->raw)
		  {
		    delta=(1000000*(now.tv_sec-cs->last_report_capacity_time.tv_sec)) +
		      (now.tv_usec-cs->last_report_capacity_time.tv_usec) ;
		    if(delta>REPORT_CAPACITY_PERIOD_USEC)
		      {
			gettimeofday(&(cs->last_report_capacity_time),NULL) ;
			report_remote_capacity(cs) ;
		      }
		  }
	      }
	    
	    cs->weight=(cs->send_capacity - cs->send_utilization) ;
	    if(cs->weight<0.0001)
	      cs->weight=0.0001 ;
	  }
    }
}

void reset_connection_state_counters(struct connection_state *cs)
{
  int i ;

  cs->total_bytes_in=0 ;
  cs->total_bytes_out=0 ;
  cs->total_too_late=0 ;
  cs->weight=0.0 ;
  cs->send_packet_history_index=0 ;
  cs->receive_packet_history_index=0 ;
  cs->send_capacity=0.0 ;
  cs->receive_capacity=0.0 ;
  cs->send_utilization=0.0 ;
  cs->receive_utilization=0.0 ;
  for(i=0;i<HISTORY_LENGTH;i++)
    {
      cs->send_packet_length_history[i]=0 ;
      cs->send_packet_time_history[i].tv_sec=0 ;
      cs->send_packet_time_history[i].tv_usec=0 ;
      cs->receive_packet_length_history[i]=0 ;
      cs->receive_packet_time_history[i].tv_sec=0 ;
      cs->receive_packet_time_history[i].tv_usec=0 ;
    }
  gettimeofday(&(cs->initialization_time),NULL) ;
  gettimeofday(&(cs->last_report_capacity_time),NULL) ;
  if(TTP_DEBUG>0)
    if(cs->debug_file!=NULL)
      fprintf(cs->debug_file,"%s reset_connection_state_counters\n",timestamps()) ;
}

struct connection_state *connection_state_alloc(char *name)
{
  struct connection_state *cs ;

  cs=(struct connection_state *)malloc(sizeof(struct connection_state)) ;
  if(cs!=NULL)
    {
      cs->name=(char *)malloc(256) ;
      sprintf(cs->name,"%s",name) ;
      cs->previous=NULL ;
      cs->pid= -1 ;
      cs->active=0 ;
      cs->standby=0 ;
      cs->dead=0 ;
      cs->cb=circular_buffer_alloc() ;
      sprintf(cs->debug_filename,"%s.debug",cs->name) ;
      cs->debug_file=NULL ;
      if(TTP_KEEP_DEBUG_LOGS)
	cs->debug_file=fopen(cs->debug_filename,"a") ;
      if(TTP_DEBUG>0)
	if(cs->debug_file!=NULL)
	  fprintf(cs->debug_file,"\n\n%s --- initialize new interface %s --- \n",timestamps(),cs->name) ;
      reset_connection_state_counters(cs) ;
    }
  return(cs) ;
}

int dup_counter=1 ;

struct connection_state *dup_connection_state(struct connection_state *x)
{
  char suffix[128] ;
  char *newname ;
  char *excl ;
  struct connection_state *cs ;

  newname=malloc(strlen(x->name)+128) ;
  sprintf(newname,"%s",x->name) ;
  excl=strstr(newname,"!") ;
  if(excl!=NULL)
    excl[0]=0 ;
  sprintf(suffix,"!%d",dup_counter++) ;
  strcat(newname,suffix) ;
  cs=connection_state_alloc(newname) ;
  free(newname) ;
  cs->interface=x->interface ;
  cs->side=x->side ;
  return(cs) ;
}

void free_connection_state(struct connection_state *x)
{
  int i ;

  if(x!=NULL)
    {
      if(x->name!=NULL)
	free(x->name) ;
      for(i=0;i<2;i++)
	{
	  close(x->ttp_to_interface[i]) ;
	  close(x->interface_to_ttp[i]) ;
	  close(x->ttp_death_signal[i]) ;
	}
      if(x->cb!=NULL)
	circular_buffer_free(x->cb) ;
      if(x->debug_file!=NULL)
	fclose(x->debug_file) ;
      free(x) ;
    }
}

/* this is a relatively weak transport encryption (typical cycle lengths are less than 1000) */
short trux_simple_rand(long *seed)
{
  short result ;

  *seed = (*seed ^ (*seed << 1)) + (*seed >> 1) ;
  result=(*seed % 0x10000) & 0xFFFF ;

  return(result) ;
}

int strmatch(char *a, char *b)
{
  int result, la, lb, n ;

  result=0 ;
  if(a==b)
    result=1 ;
  if((a!=NULL)&&(b!=NULL))
    {
      la=strlen(a) ;
      lb=strlen(b) ;
      if(la<lb)
	n=la ;
      else
	n=lb ;
      result=!strncmp(a,b,n) ;
    }
  return(result) ;
}

int syswrap_read(int fd, char *b, int n)
{
  int result,total_result,read_done=0 ;
  
  total_result=0 ;
  for(;!read_done;)
    {
      read_done=1 ;
      result=read(fd,b,n) ;
      if(result>=0)
	total_result+=result ;
      else
	{
	  if(errno==EINTR)
	    read_done=0 ;
	  else
	    total_result=result ;
	}
    }
  return(total_result) ;
}

int syswrap_write(int fd, char *b, int n)
{
  int result,total_result,write_done=0 ;

  total_result=0 ;
  for(;!write_done;)
    {
      write_done=1 ;
      result=write(fd,b,n) ;
      if(result>=0)
	total_result+=result ;
      else
	{
	  if(errno==EINTR)
	    write_done=0 ;
	  else if(errno==EAGAIN)
	    write_done=1 ;
	  else
	    total_result=result ;
	}
    }
  return(total_result) ;
}

int sure_write(long *output_seed,int fd,char *b_p,int n,int max_block_size)
{
  int result, total_result, i, block_size ;
  char *b ;

  total_result=0 ;
  if(output_seed==NULL)
    b=b_p ;
  else
    {
      /* encrypt output */
      b=(char *)malloc(n) ;
      memcpy(b,b_p,n) ;
      for(i=0;i<n;i++)
	b[i]^=(0xFF & trux_simple_rand(output_seed)) ;
    }

  for(i=0;i<n;)
    {
      block_size=n-i ;
      if(max_block_size>0)
	if(block_size>max_block_size)
	  block_size=max_block_size ;
      result=syswrap_write(fd,&(b[i]),block_size) ;
      /* result=send(fd,&(b[i]),n-i,0) ; */
      if(result<0)
        {
          perror("sure_write: write") ;
          i=n ;  /* escape from loop */
        }
      else if(result==0)  /* would block */
	{
	  fprintf(stderr,"backing off of blocking write (dropping data) (fd=%d, n=%d)\n",fd,n) ;
          i=n ;  /* escape from loop */
	}
      else
	{
	  i+=result ;
	  total_result+=result ;
	}
    }

  if(b!=b_p)
    free(b) ;
  return(total_result) ;
}

void sure_ascii_write(long *output_seed,int fd,char *s)
{
  sure_write(output_seed,fd,s,strlen(s),0) ;
}

static int sure_read(long *input_seed,int fd,char *b,int n,struct timeval *timeout)
{
  int result ;
  int i, j, virgin ;
  fd_set readfds, writefds, exceptfds ;
  struct timeval starttime ;
  struct timeval currenttime ;
  struct timeval newtimeout ;
  
  if(timeout!=NULL)
    gettimeofday(&starttime,NULL) ;
  virgin=1 ;
  for(i=0;i<n;)
    {
      if(timeout!=NULL)
	{
	  gettimeofday(&currenttime,NULL) ;
	  newtimeout.tv_sec=timeout->tv_sec-(currenttime.tv_sec-starttime.tv_sec) ;
	  newtimeout.tv_usec=timeout->tv_usec-(currenttime.tv_usec-starttime.tv_usec) ;
	  for(;newtimeout.tv_usec<0;)
	    {
	      newtimeout.tv_usec+=1000000 ;
	      newtimeout.tv_sec-=1 ;
	    }
	  if(newtimeout.tv_sec<0)
	    if(virgin)
	      {
		newtimeout.tv_sec=0 ;
		newtimeout.tv_usec=0 ;
	      }
	    else
	      return(i) ;
	  virgin=0 ;

	  FD_ZERO(&readfds) ;
	  FD_SET(fd,&readfds) ;
	  result=select(fd+1,&readfds,NULL,NULL,&newtimeout) ;
	  if(result== -1)
	    {
	      printf("sure_read: select\n") ;
	      if(errno==EBADF)
		return(i) ;
	    }
	  if(result>0 && FD_ISSET(fd,&readfds))
	    result=syswrap_read(fd,&(b[i]),n-i) ;
	  else
	    result=0 ;
	}
      else
	result=syswrap_read(fd,&(b[i]),n-i) ;
      if(result<0)
	{
	  printf("sure_read: read\n") ;
          i=n ;  /* escape from loop */
	}
      else
        {
	  /* decrypt input */
	  if(input_seed!=NULL)
	    for(j=i;j<i+result;j++)
	      b[j]^=(0xFF & trux_simple_rand(input_seed)) ;
          i+=result ;
        }
    }
  return(n) ;
}

#include "ttp_ppp.c"
#include "ttp_tty.c"
#include "ttp_tcp.c"
#include "ttp_udp.c"

void create_ttp_header(int magic, char *header,char *buf,int buflen,int position)
{
  int checksum ;
  int i ;
  struct timeval t ;

  gettimeofday(&t,NULL) ;
  checksum=(71*position)+(57*buflen) ;
  if(buf!=NULL)
    for(i=0;i<buflen;i++)
      checksum+=(i+1)*buf[i] ;

  (((int *)header)[0])=htonl(magic) ;
  (((int *)header)[1])=htonl(checksum) ;
  (((int *)header)[2])=htonl(buflen) ;
  (((int *)header)[3])=htonl(position) ;
  (((int *)header)[4])=htonl(t.tv_sec) ;
  (((int *)header)[5])=htonl(t.tv_usec) ;
}

/* returns nonzero if send was aborted (would block) */
int send_data_to_interface(struct connection_state *cs, char *buf, int buflen, int position, int magic)
{
  char packet[TTP_HEADER_SIZE+MAX_TTP_PACKET_SIZE] ;
  int send_aborted=0 ;
  int r, checksum ;

  if(cs->raw)  /* write raw data */
    {
      if(TTP_DEBUG>1)
	if(cs->debug_file!=NULL)
	  fprintf(cs->debug_file,"%s sending %d bytes of raw data at pos %d\n",timestamps(),buflen,position) ;
      if(magic==TTP_MAGIC)
	{
	  r=sure_write(NULL,cs->ttp_to_interface[1],buf,buflen,0) ;
	  if(r!=buflen)
	    send_aborted=1 ;
	}
      else
	{
	  fprintf(stderr,"ERROR: cannot send magic=0x%x over raw interface (%s)\n",magic,cs->name) ;
	}
    }
  else
    {
      create_ttp_header(magic,packet,buf,buflen,position) ;
      if((magic==TTP_MAGIC)||(magic==TTP_FEEDBACK))
	{	
	  checksum=ntohl(((int *)packet)[1]) ;
	  if(TTP_DEBUG>1)
	    if(cs->debug_file!=NULL)
	      fprintf(cs->debug_file,"%s sending packet with checksum 0x%X (len=%d, pos=%d) to %s\n",timestamps(),checksum,buflen,position,cs->name) ;
	  memmove(&(packet[TTP_HEADER_SIZE]),buf,buflen) ;
	  buflen+=TTP_HEADER_SIZE ;
	}
      if(magic==TTP_RETRANSMIT)
	{
	  buflen=TTP_HEADER_SIZE ;
	}
      r=sure_write(NULL,cs->ttp_to_interface[1],packet,buflen,0) ;
      if((magic==TTP_MAGIC)||(magic==TTP_FEEDBACK))
	if(r!=buflen)
	  send_aborted=1 ;
    }
  gettimeofday(&(cs->last_send_time),NULL) ;
  cs->total_bytes_out+=buflen ;
  
  gettimeofday(&(cs->send_packet_time_history[cs->send_packet_history_index]),NULL) ;
  cs->send_packet_length_history[cs->send_packet_history_index]=buflen ;
  cs->send_packet_history_index++ ;
  cs->send_packet_history_index%=HISTORY_LENGTH ;

  return(send_aborted) ;
}

/* for debugging ... */
int disburse_counter=1 ;

/* returns number of bytes disbursed */
int disburse_data(int magic,char *buf,int buflen,int position,int side,int spam_all_channels)
{
  struct connection_state *cs ;
  int i, r, num_choices ;
  int send_aborted, result ;
  double weight, total_weight, dice ;
  int verbose=0 ;
  int found=0 ;
  int index, x ;

  disburse_counter++ ;
  disburse_counter%=8192 ;
  if(disburse_counter==0)
    verbose=1 ;

  if(verbose)
    fprintf(stderr,"ttp: disburse %d bytes at position %d to side %d ... (disburse counter %d)\n",buflen,position,side,disburse_counter) ;

  send_aborted=0 ;
  if(buflen>0)
    {
      num_choices=0 ;
      for(cs=connections;cs!=NULL;cs=cs->previous)
	{
	  cs->destination_flag=0 ;
	  if((cs->active)&&(!(cs->standby)))
	    if(cs->side==side)
	      num_choices++ ;
	}
      if(verbose)
	fprintf(stderr,"ttp_disburse: num choices = %d\n",num_choices) ;
      if(num_choices>0)
	{
	  recalculate_weights(verbose) ;
	  total_weight=0.0 ;

	  x=0 ;
	  /* choose any underutilized channels */
	  for(cs=connections;cs!=NULL;cs=cs->previous)
	    if((cs->active)&&(!(cs->standby)))
	      if(cs->side==side)
		{
		  total_weight+=cs->weight ;
		  if(((cs->send_utilization + cs->receive_utilization) <
		      (cs->send_capacity * UNDERUTILIZATION_THRESHHOLD)) ||
		     (spam_all_channels))
		    {
		      cs->destination_flag=1 ;
		      found++ ;
		      index=x ;
		    }
		  x++ ;
		}

	  if(!found)
	  /* roll the dice */
	    {
	      x=0 ;
	      dice=drand48()*total_weight ;
	      for(cs=connections;(cs!=NULL)&&(!found);cs=cs->previous)
		if((cs->active)&&(!(cs->standby)))
		  if(cs->side==side)
		    {
		      dice-=cs->weight ;
		      if(dice<=0.0)
			{
			  cs->destination_flag=1 ;
			  found=1 ;
			  index=x ;
			}
		      x++ ;
		    }
		}
	}
    }

  if(!found)
    {
      if(magic==TTP_MAGIC)
	fprintf(stderr,"ttp_disburse: data dropped for side %d (len=%d, pos=%d)\n",side,buflen,position) ;
    }
  else
    {
      send_aborted=1 ;
      for(cs=connections;cs!=NULL;cs=cs->previous)
	if(cs->destination_flag)
	  {
	    r=send_data_to_interface(cs,buf,buflen,position,magic) ;
	    if(!r)
	      send_aborted=0 ;
	    else
	      {
		found-- ;
		if(TTP_DEBUG>0)
		  if(cs->debug_file!=NULL)
		    {
		      fprintf(cs->debug_file,"%s disburse_data aborted\n",timestamps()) ;
		      cs->send_capacity/=2.0 ;
		      cs->weight=0.0 ;
		    }
	      }
	  }
      if(found==1)
	{
	  fprintf(stderr,"%c",'A'+index) ;
	}
      else
	fprintf(stderr,"%d",found) ;
    }
  if(send_aborted)
    {
      if(!spam_all_channels)
	result=disburse_data(magic,buf,buflen,position,side,1) ;
      else
	{
	  fprintf(stderr,"disburse_data aborted\n") ;
	  result=0 ;
	}
    }
  else
    result=buflen ;
  return(result) ;
}

void ttp_shift_buffer(int side, int delta)
{
  int x=side ;
  int buflen=delta ;
  double r, f1, f2 ;

  if(buflen>STATE_BUFFER_SIZE)
    buflen=STATE_BUFFER_SIZE ;

  memmove(&(history_buffer[x][0]),&(history_buffer[x][buflen]),STATE_BUFFER_SIZE-buflen) ;  /* shift out old history data */
  memmove(&(history_buffer[x][STATE_BUFFER_SIZE-buflen]),&(side_buffer[x][0]),buflen) ;  /* shift out used data into history buffer*/

  memmove(&(side_buffer[x][0]),&(side_buffer[x][buflen]),STATE_BUFFER_SIZE-buflen) ;  /* shift out used data */
  memmove(&(side_flags[x][0]),&(side_flags[x][buflen]),STATE_BUFFER_SIZE-buflen) ;  /* shift out used flags */
  memset(&(side_flags[x][STATE_BUFFER_SIZE-buflen]),0,buflen) ;  /* invalidate new data area */

  f1=((double) side_counter[x]) ;
  f2=((double) side_counter[x]+delta) ;
  r=(double) ROLLOVER ;
  f1=floor(f1/r) ;
  f2=floor(f2/r) ;

  side_counter[x]+=delta ;
  side_counter[x]%=ROLLOVER ;

  if(f1!=f2)
    {
      /* this condition means that the counter wraps (modulo ROLLOVER) */
      fprintf(stderr,"side %d goes from %g to %g (bytes == %d)\n",side,f1,f2,side_counter[x]) ;
    }
}

void ttp_retransmit(position, length, side)
{
  int ix, delta ;

  ix=side_counter[side]-position ;
  if(ix<0)
    ix+=ROLLOVER ;
  if(ix>STATE_BUFFER_SIZE)
    {
      /* asking for too ancient history */
      delta=ix-STATE_BUFFER_SIZE ;
      ix=STATE_BUFFER_SIZE ;
      length-=delta ;
    }
  if(length>0)
    {
      if(length>STATE_BUFFER_SIZE)
	length=STATE_BUFFER_SIZE ;
      position=side_counter[side]-ix ;
      ix=STATE_BUFFER_SIZE-ix ;
      /* fprintf(cs->debug_file,"%s : retransmitting %d bytes at position %d to side %d\n",timestamps(),length,position,side) ; */
      fprintf(stderr,"r") ;
      disburse_data(TTP_MAGIC,&(history_buffer[side][ix]),length,position,side,1) ;
    }
}

void ttp_check_output_buffer(int side)
{
  int packet_sent, buflen, position, fvdix, r ;
  long delta,retransmit_delta ;
  struct timeval now ;
  char *q ;

  /* fprintf(stderr,"ttp: check output buffer for side %d -- total bytes == %d\n",side,side_total[side]) ; */
  if(side_total[side]>0)
    {
      if(!((side_flags[side][0])&VALID_FLAG))
	{
	  for(fvdix=0;(fvdix<STATE_BUFFER_SIZE)&&(!((side_flags[side][0])&VALID_FLAG));fvdix++) ;
	  gettimeofday(&now,NULL) ;
	  delta=1000000*(now.tv_sec-oldest_packet_time[side].tv_sec) + (now.tv_usec-oldest_packet_time[side].tv_usec) ;
	  /* fprintf(stderr,"ttp: check abandon delta = %d for side %d\n",delta,side) ; */
	  if(delta>ABANDON_TIMEOUT_USEC)
	    {
	      /* find amount of abandoned data */
	      for(buflen=0;((!((side_flags[side][buflen])&VALID_FLAG)) &&
			    (buflen<MAX_TTP_PACKET_SIZE));buflen++) ;
	      fprintf(stderr,"_") ;
	      /* fprintf(stderr,"%s abandon %d bytes of missing data at pos %d for side %d (delta=%ld, old pos=%d) total=%d, first valid data ix=%d\n",timestamps(),buflen,side_counter[side],side,delta,oldest_packet_position[side],side_total[side],fvdix) ; */
	      total_abandoned[side]+=buflen ;
	      ttp_shift_buffer(side,buflen) ;
	      /* fprintf(stderr,"\tnew pos %d, total=%d\n",side_counter[side],side_total[side]) ; */
	    }
	  else
	    {
	      q="q" ;
	      if(delta>RETRANSMIT_TIMEOUT_USEC)
		{
		  int new_missing=0 ;
		  int first_missing=0 ;

		  retransmit_delta=1000000*(now.tv_sec-last_retransmit_time[side].tv_sec) + (now.tv_usec-last_retransmit_time[side].tv_usec) ;
		  if(retransmit_delta>RETRANSMIT_POLL_PERIOUD_USEC)
		    {
		      for(buflen=0;((!((side_flags[side][buflen])&VALID_FLAG)) &&
				    (buflen<MAX_TTP_PACKET_SIZE));buflen++)
			if((side_flags[side][buflen])&RETRANSMIT_FLAG)
			  {
			    side_flags[side][buflen]&=~RETRANSMIT_FLAG ;
			    q="Q" ; 
			  }
		    }
		  /* find amount of missing data */
		  for(buflen=0;((!((side_flags[side][buflen])&VALID_FLAG)) &&
				(buflen<MAX_TTP_PACKET_SIZE));buflen++)
		    if(!((side_flags[side][buflen])&RETRANSMIT_FLAG))
		      {
			side_flags[side][buflen] |= RETRANSMIT_FLAG ;
			if(new_missing==0)
			  first_missing=side_counter[side]+buflen ;
			new_missing++ ;
		      }
		  if(new_missing>0)
		    {
		      fprintf(stderr,"%s",q) ;
		      /* fprintf(stderr,"%s request retransmit %d bytes of missing data at pos %d for side %d (delta=%ld, old pos=%d) total=%d\n",timestamps(),buflen,side_counter[side],side,delta,oldest_packet_position[side],side_total[side]) ; */
		      /* request retransmit */
		      disburse_data(TTP_RETRANSMIT,NULL,buflen,first_missing,!side,1) ;
		      gettimeofday(&(last_retransmit_time[side]),NULL) ;
		    }
		}
	    }
	}

      for(packet_sent=1;packet_sent;)
	{
	  packet_sent=0 ;
	  /* consume first valid part of buffer */
	  for(buflen=0;((side_flags[side][buflen])&VALID_FLAG)&&(buflen<MAX_TTP_PACKET_SIZE);buflen++) ;
	  if(buflen>0)
	    {
	      position=side_counter[side] ;
	      r=disburse_data(TTP_MAGIC,&(side_buffer[side][0]),buflen,position,side,0) ;
	      if(r>0)
		{
		  ttp_shift_buffer(side,buflen) ;
		  gettimeofday(&(oldest_packet_time[side]),NULL) ;
		  oldest_packet_position[side]=position+buflen ;
		  side_total[side]-=buflen ;
		  packet_sent=1 ;
		}
	    }
	}

      /* this block is for debugging */
      if(side_total[side]>0)
	{
	  int first_flag ;
	  for(first_flag=0;first_flag<MAX_TTP_PACKET_SIZE;first_flag++)
	    if(side_flags[side][first_flag]&VALID_FLAG)
	      {
		/* fprintf(stderr,"%s leave COB side %d, first flag=%d, total=%d, pos0=%d\n",timestamps(),side,first_flag,side_total[side],side_counter[side]) ; */
		first_flag=MAX_TTP_PACKET_SIZE ;
	      }
	}
    }
}

int baseline_delta=2048000000 ;

void ttp_process_packet(char *buf, int side, struct connection_state *cs)
{
  int i, j, x, eos, checksum, buflen, position, delta, ignore ;
  char header[TTP_HEADER_SIZE] ;
  struct timeval sent_t, received_t ;

  checksum=ntohl(((int *)buf)[1]) ;
  buflen=ntohl(((int *)buf)[2]) ;
  position=ntohl(((int *)buf)[3]) ;
  
  sent_t.tv_sec=ntohl(((int *)buf)[4]) ;
  sent_t.tv_usec=ntohl(((int *)buf)[5]) ;

  gettimeofday(&received_t,NULL) ;

  delta=(received_t.tv_sec - sent_t.tv_sec) ;
  if(delta< -2048)
    delta= -2048 ;
  if(delta> 2048)
    delta= 2048 ;
  delta*=1000000 ;
  delta+=(received_t.tv_usec - sent_t.tv_usec) ;
  if(delta<baseline_delta)
    {
      baseline_delta=delta ;
    }
  delta-=baseline_delta ;

  if(TTP_DEBUG>1)
    if(cs->debug_file!=NULL)
      {
	fprintf(cs->debug_file,"%s ttp: processing packet with checksum 0x%X of size=%d, pos=%d, delta=%d\n",timestamps(),checksum,buflen,position,delta) ;
	fprintf(cs->debug_file,"\t sent (%ld %ld), rcvd (%ld %ld)\n",sent_t.tv_sec,sent_t.tv_usec,received_t.tv_sec,received_t.tv_usec) ;
      }

  /*
    fprintf(stderr,"%s ttp: processing packet with checksum 0x%X of size=%d, pos=%d, delta=%d\n",timestamps(),checksum,buflen,position,delta) ;
    fprintf(stderr,"\t sent (%d %d), rcvd (%d %d)\n",sent_t.tv_sec,sent_t.tv_usec,received_t.tv_sec,received_t.tv_usec) ;
  */

  create_ttp_header(TTP_FEEDBACK,header,&(buf[TTP_HEADER_SIZE]),sizeof(double),0) ;
  if(!memcmp(header,buf,TTP_CORE_HEADER_SIZE))
    {
      memcpy(&(cs->send_capacity),&(buf[TTP_HEADER_SIZE]),sizeof(double)) ;
      if(TTP_DEBUG>1)
	fprintf(cs->debug_file,"%s received send capacity (%g) for %s\n",timestamps(),cs->send_capacity,cs->name) ;
    }
  else
    {
      create_ttp_header(TTP_MAGIC,header,&(buf[TTP_HEADER_SIZE]),buflen,position) ;
      if(memcmp(header,buf,TTP_CORE_HEADER_SIZE))
	{
	  fprintf(stderr,"packet with bad checksum 0x%X dropped (len=%d, pos=%d)\n",checksum,buflen,position) ;
	  if(TTP_DEBUG>0)
	    if(cs->debug_file!=NULL)
	      fprintf(cs->debug_file,"%s packet with bad checksum 0x%X dropped (len=%d, pos=%d)\n",timestamps(),checksum,buflen,position) ;
	}
      else
	{
	  /* destination is opposite originating side */
	  if(side)
	    x=0 ;  /* left */
	  else
	    x=1 ;  /* right */
	  if(ttp_is_relay)
	    {
	      disburse_data(TTP_MAGIC,&(buf[TTP_HEADER_SIZE]),buflen,position,x,0) ;
	      position+=buflen ;
	      position%=ROLLOVER ;
	      side_counter[x]=position ;
	    }
	  else
	    {
	      /* assemble packet into left or right buffer */
	      delta=position-side_counter[x] ;
	      if(delta<((0-ROLLOVER)/2))
		delta+=ROLLOVER ;
	      if(delta<0)
		{
		  fprintf(stderr,"d") ;
		  /* fprintf(stderr,"%s dropping abandoned data which finally arrived for side %d (len=%d, pos=%d) from [%s]\n",timestamps(),x,buflen,position,cs->name) ; */
		  cs->total_too_late+=buflen ;
		}
	      else
		{
		  if(delta+buflen>STATE_BUFFER_SIZE)  /* overrun buffer size */
		    {
		      delta=(delta+buflen)-STATE_BUFFER_SIZE ;
		      /* salvage old incomplete fragments */
		      for(i=0;(i<delta)&&(i<STATE_BUFFER_SIZE);i++)
			if((side_flags[x][i])&VALID_FLAG)
			  {
			    eos=0 ;
			    for(j=i;(!eos)&&(j<delta)&&(j<STATE_BUFFER_SIZE);)
			      if((side_flags[x][j])&VALID_FLAG)
				j++ ;
			      else
				eos=1 ;
			    fprintf(stderr,"buffer will overrun -- salvage %d early bytes for side %d\n",j-i,x) ;
			    disburse_data(TTP_MAGIC,&(side_buffer[x][i]),j-i,(side_counter[x]+i)%ROLLOVER,x,0) ;
			    side_total[x]-=(j-i) ;
			    i=j ;
			  }
		      ttp_shift_buffer(x,delta) ;
		      delta=position-side_counter[x] ;  /* new delta is within buffer limits */
		      if(delta<0)
			delta+=ROLLOVER ;
		    }
		  
		  /*
		    fprintf(stderr,"assemble %d bytes into buffer pos %d (offset %d) for side %d\n",buflen,position,delta,x) ;
		  */
		  
		  /* copy new buffer piece into place */
		  memcpy(&(side_buffer[x][delta]),&(buf[TTP_HEADER_SIZE]),buflen) ;
		  ignore=0 ;
		  for(i=0;i<buflen;i++)
		    {
		      if(side_flags[x][delta+i]&VALID_FLAG)
			ignore++ ;
		      side_flags[x][delta+i]=VALID_FLAG ;
		    }
		  
		  if(ignore>0)
		    {
		      cs->total_too_late+=ignore ;
		      /* fprintf(stderr,"ignoring %d redundant bytes in packet of size %d for side %d\n",ignore,buflen,x) ; */
		      fprintf(stderr,"i") ;
		    }
		  
		  if(side_total[x]==0)  /* we're the first packet put in the buffer */
		    {
		      gettimeofday(&(oldest_packet_time[x]),NULL) ;
		      oldest_packet_position[x]=position ;
		    }
		  side_total[x]+=buflen-ignore ;
		}
	      /* ttp_check_output_buffer(x) ; */
	    }
	}
    }
}

void signal_death(struct connection_state *cs)
{
  char buf[64] ;
  switch(cs->interface)
    {
    case ttp_tcp_transient:
      fprintf(stderr,"signalling death of interface %s...\n",cs->name) ;
      ((int *)buf)[0]=htonl(TTP_PING) ;
      sure_write(NULL,cs->ttp_death_signal[1],buf,sizeof(int),0) ;
      break ;;;
    case ttp_tcp_client:
    case ttp_udp:
      fprintf(stderr,"marking interface %s as dead...\n",cs->name) ;
      cs->dead=1 ;
      reload_flag=1 ;
      break ;;;
    default:
      fprintf(stderr,"interface %s resists death...\n",cs->name) ;
    }
}

void load_and_parse_resource_file() ;  /* prototype */

void ttp()
{
  struct connection_state *cs, *os ;
  int done=0 ;
  int i, r, x, maxfd, buflen, got_packet ;
  fd_set readfds ;
  char buf[TTP_HEADER_SIZE+MAX_TTP_PACKET_SIZE] ;
  struct timeval timeout ;
  struct timeval last_status_line_time ;  /* when did we last print a status line? */
  struct timeval now ;
  struct sockaddr_in remaddr ;
  int magic, length, position, garbage_length, standby, opposite_side_ready ;
  unsigned long rs ;
  long littlest_delta, delta ;
  char *b ;

  gettimeofday(&now,NULL) ;
  gettimeofday(&last_status_line_time,NULL) ;
  rs=((unsigned long) getpid())*now.tv_usec + now.tv_sec ;
  rs^=(unsigned long) &rs ;
  srand48(rs) ;

  /* initialize */
  for(x=0;x<2;x++)
    {
      side_counter[x]=0 ;
      side_total[x]=0 ;
      total_abandoned[x]=0 ;
      for(i=0;i<STATE_BUFFER_SIZE;i++)
	side_flags[x][i]=0 ;
    }

  for(;!done;)
    {
      if(reload_flag)
	{
	  reload_flag=0 ;
	  load_and_parse_resource_file() ;
	}

      maxfd= -1 ;
      FD_ZERO(&readfds) ;
      for(cs=connections;cs!=NULL;cs=cs->previous)
        {
          FD_SET(cs->interface_to_ttp[0],&readfds) ;
          if(cs->interface_to_ttp[0]>maxfd)
            maxfd=cs->interface_to_ttp[0] ;
        }

      littlest_delta=PING_PERIOD_USEC ;
      for(x=0;x<2;x++)
	if(side_total[x]>0)
	  {
	    gettimeofday(&now,NULL) ;
	    delta=ABANDON_TIMEOUT_USEC ;
	    delta-=(1000000*(now.tv_sec-oldest_packet_time[x].tv_sec)) + (now.tv_usec-oldest_packet_time[x].tv_usec) ;
	    if(delta<littlest_delta)
	      littlest_delta=delta ;
	  }
      if(littlest_delta>0)
	{
	  timeout.tv_sec=littlest_delta/1000000 ;
	  timeout.tv_usec=littlest_delta%1000000 ;
	}
      else
	{
	  timeout.tv_sec=0 ;
	  timeout.tv_usec=0 ;
	}

      /* fprintf(stderr,"ttp: selecting %d secs, %d msecs...\n",timeout.tv_sec,timeout.tv_usec/1000) ; */
      r=select(maxfd+1,&readfds,NULL,NULL,&timeout) ;

      if(r>0)
	for(cs=connections;cs!=NULL;cs=cs->previous)
	  if(FD_ISSET(cs->interface_to_ttp[0],&readfds))
	    {
	      /* fprintf(stderr,"ttp: interface %s selected\n",cs->name) ; */
	      if((buflen=syswrap_read(cs->interface_to_ttp[0],&(buf[TTP_HEADER_SIZE]),MAX_TTP_PACKET_SIZE))>0)
		{
		  cs->total_bytes_in+=buflen ;
		  /* fprintf(stderr,"ttp: read %d bytes from interface %s (total %ld)\n",buflen,cs->name,cs->total_bytes_in) ; */
		  if(TTP_DEBUG>1)
		    if(cs->debug_file!=NULL)
		      fprintf(cs->debug_file,"%s: read %d bytes (total so far %ld)\n",timestamps(),buflen,cs->total_bytes_in) ;
		  if(cs->raw)  /* raw data */
		    {
		      /* encapsulate */
		      position=side_counter[1-(cs->side)] ;
		      create_ttp_header(TTP_MAGIC,buf,&(buf[TTP_HEADER_SIZE]),buflen,position) ;
		      ttp_process_packet(buf,cs->side,cs) ;
		    }
		  else  /* examine ttp packet */
		    { 
		      gettimeofday(&(cs->newest_packet_time),NULL) ;
		      gettimeofday(&(cs->receive_packet_time_history[cs->receive_packet_history_index]),NULL) ;
		      cs->receive_packet_length_history[cs->receive_packet_history_index]=buflen ;
		      cs->receive_packet_history_index++ ;
		      cs->receive_packet_history_index%=HISTORY_LENGTH ;

		      r=circular_buffer_write(cs->cb,&(buf[TTP_HEADER_SIZE]),buflen) ;
		      /* fprintf(stderr,"cb write %d bytes gives error result %d (cb->current_length=%d)\n",buflen,r,cs->cb->current_length) ; */
		      for(got_packet=1;got_packet;)
			{
			  got_packet=0 ;
			  garbage_length=0 ;
			  for(r=1;r==1;)
			    {
			      r=0 ;
			      if(circular_buffer_peek(cs->cb,buf,sizeof(int)))
				{
				  /* fprintf(stderr,"cb peek (%d bytes) failed\n",sizeof(int)) ; */
				}
			      else
				{
				  magic=ntohl(((int *)buf)[0]) ;
				  if((magic==TTP_MAGIC) || (magic==TTP_RETRANSMIT) || (magic==TTP_FEEDBACK))
				    {
				      /* fprintf(stderr,"got magic number from interface %s (cb->current_length=%d)\n",cs->name,cs->cb->current_length) ; */
				      if(!circular_buffer_peek(cs->cb,buf,TTP_HEADER_SIZE))  /* if successful peek */
					r=TTP_HEADER_SIZE ;
				    }
				  else
				    {
				      if(magic==TTP_PING)
					{
					  r=1 ;
					  circular_buffer_read(cs->cb,buf,sizeof(int)) ;
					}
				      else if(magic==TTP_ACTIVE)
					{
					  r=1 ;
					  circular_buffer_read(cs->cb,buf,sizeof(int)) ;
					  fprintf(stderr,"%s: interface (%s) becomes active\n",timestamps(),cs->name) ;
					  gettimeofday(&(cs->last_send_time),NULL) ;
					  cs->active=1 ;
					}
				      else if(magic==TTP_INACTIVE)
					{
					  r=1 ;
					  circular_buffer_read(cs->cb,buf,sizeof(int)) ;
					  fprintf(stderr,"%s: interface (%s) becomes inactive\n",timestamps(),cs->name) ;
					  cs->active=0 ;
					}
				      else if(!circular_buffer_read(cs->cb,buf,1))  /* dispose non-packet data */
					{
					  r=1 ;
					  garbage_length++ ;
					  if(garbage_length<8)
					    fprintf(stderr,"dispose garbage char 0x%X (%c) from %s\n",buf[0],buf[0],cs->name) ;
					  if(garbage_length==8)
					    fprintf(stderr,"...\n") ;
					}
				    }
				}
			    }
			  if(garbage_length>0)
			    {
			      fprintf(stderr,"ttp disposed of %d garbage characters from %s\n",garbage_length,cs->name) ;
			      signal_death(cs) ;
			    }

			  if(r==TTP_HEADER_SIZE)
			    {
			      if(magic==TTP_RETRANSMIT)
				{
				  if(circular_buffer_read(cs->cb,buf,TTP_HEADER_SIZE))
				    {
				      perror("ttp: cb read failed (1)") ;
				      exit(-1) ;
				    }
				  length=ntohl(((int *)buf)[2]) ;
				  position=ntohl(((int *)buf)[3]) ;
				  ttp_retransmit(position, length, cs->side) ;
				  got_packet=1 ;
				}
			      if((magic==TTP_MAGIC) || (magic==TTP_FEEDBACK))
				{
				  /* fprintf(stderr,"ttp: got header from interface %s, cblen=%d\n",cs->name,cs->cb->current_length) ; */
				  length=ntohl(((int *)buf)[2]) ;
				  if((length<=0)||(length>MAX_TTP_PACKET_SIZE))
				    {
				      circular_buffer_read(cs->cb,buf,1) ;  /* dispose garbage header */
				    }
				  else
				    if(cs->cb->current_length>=TTP_HEADER_SIZE+length)
				      {
					if(circular_buffer_read(cs->cb,buf,TTP_HEADER_SIZE+length))
					  {
					    perror("ttp: cb read failed (2)") ;
					    exit(-1) ;
					  }
					ttp_process_packet(buf,cs->side,cs) ;
					got_packet=1 ;
				      }
				}
			    }
			}
		    }
		}
	    }
      for(x=0;x<2;x++)
	ttp_check_output_buffer(x) ;

      gettimeofday(&now,NULL) ;

      delta=((1000000*(now.tv_sec-last_status_line_time.tv_sec)) +
	     (now.tv_usec-last_status_line_time.tv_usec)) ;
      if(delta>STATUS_LINE_PERIOD_USEC)
	{
	  gettimeofday(&last_status_line_time,NULL) ;
	  recalculate_weights(0) ;
	  print_status_line() ;
	}

      for(cs=connections;cs!=NULL;cs=cs->previous)
	if(cs->active)
	  if(!(cs->raw))
	    {
	      /* make sure remote side knows we're alive */
	      delta=((1000000*(now.tv_sec-cs->last_send_time.tv_sec)) +
		     (now.tv_usec-cs->last_send_time.tv_usec)) ;
	      if(delta>PING_PERIOD_USEC)
		{
		  opposite_side_ready=0 ;
		  for(os=connections;(!opposite_side_ready)&&(os!=NULL);os=os->previous)
		    if((os->active)&&(!(os->standby)))
		      if(os->side != cs->side)
			opposite_side_ready=1 ;
		  if(opposite_side_ready)
		    {
		      /* fprintf(stderr,"send ping to [%s]\n",cs->name) ; */
		      ((int *)buf)[0]=htonl(TTP_PING) ;
		      sure_write(NULL,cs->ttp_to_interface[1],buf,sizeof(int),0) ;
		    }
		  else
		    {
		      if(ttp_is_relay)
			fprintf(stderr,"ttp relay: side %d is not ready\n",1-cs->side) ;
		    }
		  cs->last_send_time.tv_sec  = now.tv_sec ;
		  cs->last_send_time.tv_usec = now.tv_usec ;
		}

	      /* make sure we know if remote side is alive */
	      delta=((1000000*(now.tv_sec-cs->newest_packet_time.tv_sec)) +
		     (now.tv_usec-cs->newest_packet_time.tv_usec)) ;
	      if(delta>DEATH_TIMEOUT_USEC)
		{
		  fprintf(stderr,"no response from interface %s for %ld usec\n",cs->name,delta) ;
		  signal_death(cs) ;
		}
	      else
		{
		  if(delta>=0)
		    standby=(delta>PING_TIMEOUT_USEC) ;
		  else
		    {
		      fprintf(stderr,"newest packet time delta wrapped! (%ld)\n",delta) ;
		      /* adjust */
		      cs->newest_packet_time.tv_sec=now.tv_sec-(1+(PING_TIMEOUT_USEC / 1000000)) ;
		      standby=cs->standby ;
		    }
		  if(standby!=cs->standby)
		    {
		      if(!standby)
			reset_connection_state_counters(cs) ;
		      
		      fprintf(stderr,"%s : interface [%s] changing to standby == %d (delta=%ld)\n",timestamps(),cs->name,standby,delta) ;
		      cs->standby=standby ;
		    }
		}
	    }
    }
}

void print_statistics()
{
  struct connection_state *cs ;
  struct timeval now ;
  time_t age ;
  double avg_speed ;
  long delta ;
  int side ;

  fprintf(stderr,"\n\nttp version %s\n",TTP_VERSION) ;
  fprintf(stderr,"interface\n  age      \ttotal in \ttotal out\ttoo late \tavg. speed\tweight\tnewest_rcv_delta\n") ;
  fprintf(stderr,"------------------------------------------------------------------\n") ;
  gettimeofday(&now,NULL) ;
  for(cs=connections;cs!=NULL;cs=cs->previous)
    {
      delta=((1000000*(now.tv_sec-cs->newest_packet_time.tv_sec)) +
	     (now.tv_usec-cs->newest_packet_time.tv_usec)) ;
      age=(now.tv_sec)-(cs->initialization_time.tv_sec) ;
      if(age>0)
	avg_speed=((double) (cs->total_bytes_in + cs->total_bytes_out)) / ((double) age) ;
      fprintf(stderr,"%s (%s, %s) fd[I(%d, %d) O(%d, %d)]:\n",cs->name,cs->active?"active":"incative",cs->standby?"standby":"active", cs->interface_to_ttp[0], cs->interface_to_ttp[1], cs->ttp_to_interface[0],  cs->ttp_to_interface[1]) ;
      fprintf(stderr,"  %ld        \t%ld        \t%ld        \t%ld        \t%g    \t%g\t%ld usec\n",age,cs->total_bytes_in,cs->total_bytes_out,cs->total_too_late,avg_speed,cs->weight,delta) ;
    }
  fprintf(stderr,"------------------------------------------------------------------\n\n") ;
  for(side=0;side<2;side++)
    {
      int fvdix, totval ;
      totval=0 ;
      for(fvdix=0;(fvdix<STATE_BUFFER_SIZE);fvdix++)
	if((side_flags[side][0])&VALID_FLAG)
	  totval++ ;
      for(fvdix=0;(fvdix<STATE_BUFFER_SIZE)&&(!((side_flags[side][0])&VALID_FLAG));fvdix++) ;
      fprintf(stderr,"side %d: %ld bytes abandoned fvdix=%d, totval=%d, side_total=%d\n",side,total_abandoned[side],fvdix,totval,side_total[side]) ;
    }
}

void hup_handler(int flag)
{
  if(getpid()==main_ttp_pid)
    {
      print_statistics() ;
      reload_flag=1 ;
    }
}

void usr1_handler(int flag)
{
  if(getpid()==main_ttp_pid)
    {
      print_statistics() ;
      TTP_DEBUG++ ;
      TTP_DEBUG%=4 ;
      fprintf(stderr,"toggling debug to %d\n",TTP_DEBUG) ;
    }
}

void child_handler(int flag)
{
  int cpid, status ;
  char *name ;
  struct connection_state *cs ;

  cpid=waitpid(-1,&status,WNOHANG) ;
  if(cpid>0)
    {
      name="unknown" ;
      for(cs=connections;cs!=NULL;cs=cs->previous)
	if(cs->pid==cpid)
	  name=cs->name ;
      fprintf(stderr,"ttp: child %d (%s) has exited\n",cpid,name) ;
    }

  signal(SIGCLD,child_handler) ;
}

int main(int argc, char **argv)
{
  main_ttp_pid=getpid() ;
  fprintf(stderr,"ttp version %s\n",TTP_VERSION) ;
  signal(SIGCLD,child_handler) ;
  signal(SIGHUP,hup_handler) ;
  signal(SIGUSR1,usr1_handler) ;
  if(argc!=2)
    {
      fprintf(stderr,"
ttp connects two endpoints through multiple bidirectional channels.
USAGE: ttp <filename>
  <filename> contains 1 line: ttp <interface> ... = <interface> ...
  <interface> consists of <method>:<address>[@<host>:<port>]
  <method> is one of \"ppp\" or \"udp\" or \"tcp\" or \"tty\"
  For ppp, <address> is a new local IP address, omit @<host>:<port>
  For udp, you must specify the remote address with @<host>:<port>
  For tcp, if @<host>:<port> is omitted then ttp will spawn a daemon
  listening for remote ttp connections on port number <address>,
  otherwise <address> is the local client port number (0 means use any port).
  * ttp can act as an equalizer, or as a relay.
  * When acting as an equalizer, ttp connects the left and right half
    of the ttp equation by distributing and collecting a data stream
    between the two sets of interfaces.  data builds up in the interfaces'
    circular buffers.  when an interface on one side of the equation has
    a complete ttp packet in its circular buffer, then that packet is
    unwrapped and placed at the corect position in the stream buffer of
    the other side of the equation.  For this, there are two buffers.
    When a contiguous block including element [0] is available in one
    of these two buffers, it is packetized and distributed to an
    interface on the same side of the ttp equation.
  * When acting as a relay, ttp simply passes along ttp packets from
    one side of the equation to the other.
  * To specify that ttp should behave as a relay instead of an equalizer,
    use a | instead of a = to seperate the two sides.
\n") ;

    }
  else
    {
      resource_filename=strdup(argv[1]) ;
      reload_flag=1 ;
      ttp() ;
    }

  return(0) ;
}

/* called at initialization and also when SIGHUP'ed (with reload_flag==1) */
void load_and_parse_resource_file()
{
  struct connection_state *cs, *ocs, *previous ;
  int i, j, is_pppd, argc, fatal=0, side=0, fd, flags ;
  char address[512] ;
  char temps[512] ;
  char *s, *q, *p, *line, *colon ;
  char **argv ;
  FILE *f ;

  /* these vars are passed as parameters to interface constructor functions */
  int local_port ;
  int remote_port ;
  char remote_hostname[512] ;
  char ipaddress[512] ;

  for(cs=connections;cs!=NULL;cs=cs->previous)
    cs->flag=0 ;

  fprintf(stderr,"ttp: loading configuration file %s\n",resource_filename) ;

  f=fopen(resource_filename,"r") ;
  if(f==NULL)
    {
      perror("ttp: fopen resoure_file") ;
      exit(-1) ;
    }

  line=(char *)malloc(MAX_NUM_INTERFACES*256) ;
  if(line==NULL)
    {
      perror("ttp: malloc line") ;
      exit(-1) ;
    }

  fscanf(f,"%[^\n]",line) ;
  fclose(f) ;

  argv=(char **)calloc(MAX_NUM_INTERFACES,sizeof(char *)) ;
  argv[0]="ttp" ;
  argc=1 ;
  for(s=line;(q=strtok(s," "))!=NULL;s=NULL)
    argv[argc++]=strdup(q) ;

  free(line) ;

  for(i=1;(i<argc)&&(!fatal);i++)
    {
      fprintf(stderr,"\t%s\n",argv[i]) ;
      if(!strcmp("=",argv[i]))
	{
	  ttp_is_relay=0 ;
	  side++ ;
	}
      else if(!strcmp("|",argv[i]))
	{
	  ttp_is_relay=1 ;
	  side++ ;
	}
      else
	{
	  int match, found=0 ;

	  cs=NULL ;
	  for(cs=connections;cs!=NULL;cs=cs->previous)
	    if(!cs->dead)
	      {
		sprintf(temps,"%s",cs->name) ;
		p=strstr(temps,"!") ;
		if(p!=NULL)
		  *p=0 ;
		if(!strcmp(argv[i],temps))
		  {
		    fprintf(stderr,"\t\tinterface [%s] already started... (pid %d)\n",cs->name,cs->pid) ;
		    found=1 ;
		    cs->flag=1 ;
		  }
	      }
	  
	  if(!found)
	    {
	      cs=connection_state_alloc(argv[i]) ;
	      cs->side=side ;
	      cs->flag=1 ;
	      cs->previous=connections ;
	      connections=cs ;

	      is_pppd=0 ;
	      if(strmatch("ppp:",argv[i]))
		{
		  is_pppd=1 ;
		  sprintf(pppdname,"pppd") ;
		}
	      if(strmatch("pppd.",argv[i]))
		{
		  is_pppd=1 ;
		  sprintf(pppdname,"%s",argv[i]) ;
		  for(j=0;(pppdname[j]!=0)&&(pppdname[j]!=':');j++) ;
		  pppdname[j]=0 ;
		}

	      if(is_pppd)
		{
		  cs->interface=ttp_ppp ;
		  cs->raw=1 ;
		  colon=strstr(argv[i],":") ;
		  strncpy(ipaddress,colon+1,512) ;
		}
	      else if(strmatch("tty",argv[i]))
		{
		  cs->interface=ttp_tty ;
		  cs->raw=1 ;
		}
	      else if((strmatch("tcp:",argv[i])) ||
		      (strmatch("udp:",argv[i])))
		{
		  cs->raw=0 ;
		  strncpy(address,argv[i]+4,512) ;
		  s=strstr(address,"@") ;
		  if(s==NULL)
		    {
		      if(strmatch("udp:",argv[i]))
			{
			  fprintf(stderr,"ttp: ERROR: udp channel must specify peer: %s\n",argv[i]) ;
			  fatal=1 ;
			}
		      else
			cs->interface=ttp_tcp_transient ;
		    }
		  else
		    {
		      if(strmatch("udp:",argv[i]))
			cs->interface=ttp_udp ;
		      else
			cs->interface=ttp_tcp_client ;
		      s[0]=0 ;
		      q=strstr(s+1,":") ;
		      remote_port=28710 ;  /* default */
		      if(q!=NULL)
			{
			  q[0]=0 ;
			  sscanf(q+1,"%d",&remote_port) ;
			}
		      strncpy(remote_hostname,s+1,512) ;
		    }
		  if(1!=sscanf(address,"%d",&local_port))
		    local_port=28710 ;  /* default */
		}
	      else
		{
		  fprintf(stderr,"ttp: ERROR: bad parameter: %s\n",argv[i]) ;
		  fatal=1 ;
		}
	      if(!fatal)
		switch(cs->interface)
		  {
		  case ttp_ppp :
		    init_ppp(cs,ipaddress) ;
		    break ;;;
		  case ttp_tty :
		    init_tty(cs) ;
		    break ;;;
		  case ttp_tcp_transient :
		    {
		      struct connection_state *transients[MAX_TCP_CONNECTIONS] ;
		      int j ;
		      
		      strcat(cs->name,"!0") ;
		      transients[0]=cs ;
		      for(j=1;j<MAX_TCP_CONNECTIONS;j++)
			{
			  transients[j]=dup_connection_state(cs) ;
			  transients[j]->flag=1 ;
			  transients[j]->previous=transients[j-1] ;
			  connections=transients[j] ;
			}
		      init_tcp_server(transients,MAX_TCP_CONNECTIONS,local_port) ;
		    }
		    break ;;;
		  case ttp_tcp_client :
		    init_tcp_client(cs,remote_hostname,remote_port,local_port) ;
		    break ;;;
		  case ttp_udp :
		    init_udp(cs,remote_hostname,remote_port,local_port) ;
		    break ;;;
		  }
	      if(!fatal)
		{
		  /* set write FDs to nonblocking */
		  fd=cs->ttp_to_interface[1] ;
		  if((flags=fcntl(fd,F_GETFL,0))== -1)
		    perror("fcntl: get flags") ;
		  flags = flags | O_NDELAY ;
		  if(fcntl(fd,F_SETFL,flags)== -1)
		    perror("fcntl: set flags") ;

		  fd=cs->ttp_death_signal[1] ;
		  if((flags=fcntl(fd,F_GETFL,0))== -1)
		    perror("fcntl: get flags") ;
		  flags = flags | O_NDELAY ;
		  if(fcntl(fd,F_SETFL,flags)== -1)
		    perror("fcntl: set flags") ;
		}
	    }
	}

    }
  if(side!=1)
    {
      fprintf(stderr,"ttp: ERROR: parameter list must have exactly one = or |\n") ;
      fatal=1 ;
    }

  /* check to make sure that there are no other interfaces on the same side of the = as ppp (or tty) */
  {
    int num_raw_interfaces=0 ;
    for(cs=connections;cs!=NULL;cs=cs->previous)
      if(cs->raw)
	num_raw_interfaces++ ;
    if(num_raw_interfaces>1)
      {
	fprintf(stderr,"ERROR: number of raw interfaces (%d) > 1\n",num_raw_interfaces) ;
      }
  }

  ocs=NULL ;
  for(cs=connections;cs!=NULL;cs=previous)
    {
      previous=cs->previous ;
      if(fatal || (!cs->flag))
	if(cs->pid>0)
	  {
	    fprintf(stderr,"killing interface: %s (pid %d)\n",cs->name,cs->pid) ;
	    if(TTP_DEBUG>0)
	      if(cs->debug_file!=NULL)
		fprintf(cs->debug_file,"%s killing interface: %s (pid %d)\n",timestamps(),cs->name,cs->pid) ;
	    kill(cs->pid,SIGQUIT) ;
	    if(ocs==NULL)
	      connections=previous ;
	    else
	      ocs->previous=previous ;
	    free_connection_state(cs) ;
	    cs=NULL ;
	  }
      if(cs!=NULL)
	ocs=cs ;
    }

  for(cs=connections;cs!=NULL;cs=cs->previous)
    if(cs->flag)
      {
	/* fprintf(stderr,"remaining interface: %s (pid %d)\n",cs->name,cs->pid) ; */
      }

  if(fatal)
    {
      fprintf(stderr,"ttp: fatal configuration error") ;
      exit(-1) ;
    }
}
